#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.models.genai.custom_model_validation import CustomModelValidation
from datarobot.models.genai.playground import Playground
from datarobot.models.use_cases.use_case import UseCase
from datarobot.models.use_cases.utils import get_use_case_id, resolve_use_cases, UseCaseLike
from datarobot.utils.pagination import unpaginate


def get_entity_id(entity: Union[Playground, str]) -> str:
    """
    Get the entity ID from the entity parameter.

    Parameters
    ----------
    entity : ApiObject or str
        May be entity ID or the entity.

    Returns
    -------
    id : str
        Entity ID
    """
    if isinstance(entity, str):
        return entity

    return entity.id


chunking_parameters_trafaret = t.Dict(
    {
        t.Key("embedding_model"): str,
        t.Key("chunking_method"): str,
        t.Key("chunk_size"): t.Int,
        t.Key("chunk_overlap_percentage"): t.Int,
        t.Key("separators"): t.List(t.String),
    }
).ignore_extra("*")

embedding_model_trafaret = t.Dict(
    {
        t.Key("embedding_model"): str,
        t.Key("description"): str,
        t.Key("max_sequence_length"): int,
        t.Key("languages"): str,
    }
).ignore_extra("*")

supported_embeddings_trafaret = t.Dict(
    {
        t.Key("embedding_models"): t.List(embedding_model_trafaret),
        t.Key("default_embedding_model"): t.String,
    }
).ignore_extra("*")

text_chunking_parameter_fields_trafaret = t.Dict(
    {
        t.Key("name"): t.String,
        t.Key("type"): t.String,
        t.Key("description"): t.String,
        t.Key("default"): t.Or(t.Int, t.List(t.String(allow_blank=True))),
        t.Key("min", optional=True): t.Or(t.Int, t.Null),
        t.Key("max", optional=True): t.Or(t.Int, t.Null),
    }
).ignore_extra("*")

text_chunking_method_trafaret = t.Dict(
    {
        t.Key("chunking_method"): t.String,
        t.Key("chunking_parameters"): t.List(text_chunking_parameter_fields_trafaret),
        t.Key("description"): t.String,
    }
).ignore_extra("*")

text_chunking_config_trafaret = t.Dict(
    {
        t.Key("embedding_model"): t.String,
        t.Key("methods"): t.List(text_chunking_method_trafaret),
        t.Key("default_method"): t.String,
    }
).ignore_extra("*")

supported_text_chunkings_trafaret = t.Dict(
    {
        t.Key("text_chunking_configs"): t.List(text_chunking_config_trafaret),
    }
).ignore_extra("*")

vector_database_trafaret = t.Dict(
    {
        t.Key("id"): t.String,
        t.Key("name"): t.String,
        t.Key("size"): t.Int,
        t.Key("use_case_id"): t.String,
        t.Key("dataset_id", optional=True, default=None): t.Or(t.Null, t.String),
        t.Key("embedding_model", optional=True, default=None): t.Or(t.Null, t.String),
        t.Key("chunking_method", optional=True, default=None): t.Or(t.Null, t.String),
        t.Key("chunk_size"): t.Int,
        t.Key("chunk_overlap_percentage"): t.Int,
        t.Key("chunks_count"): t.Int,
        t.Key("separators"): t.List(t.String(allow_blank=True)),
        t.Key("creation_date"): t.String,
        t.Key("creation_user_id"): t.String,
        t.Key("organization_id"): t.String,
        t.Key("tenant_id"): t.String,
        t.Key("last_update_date"): t.String,
        t.Key("execution_status"): t.String,
        t.Key("playgrounds_count"): t.Int,
        t.Key("dataset_name"): t.String(allow_blank=True),
        t.Key("user_name"): t.String,
        t.Key("source"): t.String,
        t.Key("validation_id", optional=True, default=None): t.Or(t.Null, t.String),
        t.Key("error_message", optional=True, default=None): t.Or(t.Null, t.String),
        t.Key("is_separator_regex"): t.Bool,
    }
).ignore_extra("*")


class ChunkingParameters(APIObject):
    """
    Parameters defining how documents are split and embedded.

    Attributes
    ----------
    embedding_model : str
        Name of the text embedding model.
        Currently supported options are listed in VectorDatabaseEmbeddingModel
        but the values can differ with different platform versions.
    chunking_method : str
        Name of the method to split dataset documents.
        Currently supported options are listed in VectorDatabaseChunkingMethod
        but the values can differ with different platform versions.
    chunk_size : int
        Size of each text chunk in number of tokens.
    chunk_overlap_percentage : int
        Overlap percentage between chunks.
    separators : list[str]
        Strings used to split documents into text chunks.
    """

    _converter = chunking_parameters_trafaret

    def __init__(
        self,
        embedding_model: str,
        chunking_method: str,
        chunk_size: int,
        chunk_overlap_percentage: int,
        separators: List[str],
    ):
        self.embedding_model = embedding_model
        self.chunking_method = chunking_method
        self.chunk_size = chunk_size
        self.chunk_overlap_percentage = chunk_overlap_percentage
        self.separators = separators


class EmbeddingModel(APIObject):
    """
    A single model for embedding text.

    Attributes
    ----------
    embedding_model : str
        Name of the text embedding model.
        Currently supported options are listed in VectorDatabaseEmbeddingModel
        but the values can differ with different platform versions.
    description : str
        Description of the embedding model.
    max_sequence_length : int
        The model's maximum number of processable input tokens.
    languages : str
        Languages supported by the model.
        Currently supported options are listed in VectorDatabaseDatasetLanguages
        but the values can differ with different platform versions.
    """

    _converter = embedding_model_trafaret

    def __init__(
        self,
        embedding_model: str,
        description: str,
        max_sequence_length: int,
        languages: List[str],
    ):
        self.embedding_model = embedding_model
        self.description = description
        self.max_sequence_length = max_sequence_length
        self.languages = languages


class SupportedEmbeddings(APIObject):
    """
    All supported embedding models including the recommended default model.

    Attributes
    ----------
    embedding_models : list[EmbeddingModel]
        All supported embedding models.
    default_embedding_model : str
        Name of the default recommended text embedding model.
        Currently supported options are listed in VectorDatabaseEmbeddingModel
        but the values can differ with different platform versions.
    """

    _converter = supported_embeddings_trafaret

    def __init__(
        self,
        embedding_models: List[Dict[str, Any]],
        default_embedding_model: str,
    ):
        self.embedding_models = [
            EmbeddingModel.from_server_data(model) for model in embedding_models
        ]
        self.default_embedding_model = default_embedding_model


class TextChunkingParameterFields(APIObject):
    """
    Text chunking parameter fields.

    Attributes
    ----------
    name : str
        Parameter name.
    type : str
        Parameter type.
        Currently supported options are listed in VectorDatabaseChunkingParameterType
        but the values can differ with different platform versions.
    description : str
        Parameter description.
    default : int or list[str]
        Parameter default value.
    min : int, optional
        Parameter minimum value.
    max : int, optional
        Parameter maximum value.
    """

    _converter = text_chunking_parameter_fields_trafaret

    def __init__(
        self,
        name: str,
        type: str,
        description: str,
        default: Union[int, List[str]],
        min: Optional[int] = None,
        max: Optional[int] = None,
    ):
        self.name = name
        self.type = type
        self.description = description
        self.default = default
        self.min = min
        self.max = max


class TextChunkingMethod(APIObject):
    """
    A single text chunking method.

    Attributes
    ----------
    chunking_method : str
        Name of the method to split dataset documents.
        Currently supported options are listed in VectorDatabaseChunkingMethod
        but the values can differ with different platform versions.
    chunking_parameters : list[TextChunkingParameterFields]
        All chunking parameters including ranges and defaults.
    description : str
        Description of the chunking method.
    """

    _converter = text_chunking_method_trafaret

    def __init__(
        self,
        chunking_method: str,
        chunking_parameters: List[Dict[str, Any]],
        description: str,
    ):
        self.chunking_method = chunking_method
        self.chunking_parameters = [
            TextChunkingParameterFields.from_server_data(parameter)
            for parameter in chunking_parameters
        ]
        self.description = description


class TextChunkingConfig(APIObject):
    """
    A single text chunking configurations.

    Attributes
    ----------
    embedding_model : str
        Name of the text embedding model.
        Currently supported options are listed in VectorDatabaseEmbeddingModel
        but the values can differ with different platform versions.
    methods : list[TextChunkingMethod]
        All text chunking methods and their supported parameters.
    default_method : str
        The default text chunking method name.
        Currently supported options are listed in VectorDatabaseChunkingMethod
        but the values can differ with different platform versions.
    """

    _converter = text_chunking_config_trafaret

    def __init__(
        self,
        embedding_model: str,
        methods: List[Dict[str, Any]],
        default_method: str,
    ):
        self.embedding_model = embedding_model
        self.methods = [TextChunkingMethod.from_server_data(method) for method in methods]
        self.default_method = default_method


class SupportedTextChunkings(APIObject):
    """
    Supported text chunking configurations which includes a set of
    recommended chunking parameters for each supported embedding model.

    Attributes
    ----------
    text_chunking_configs
        All supported text chunking configurations.
    """

    _converter = supported_text_chunkings_trafaret

    def __init__(self, text_chunking_configs: List[Dict[str, Any]]):
        self.text_chunking_configs = [
            TextChunkingConfig.from_server_data(config) for config in text_chunking_configs
        ]


class VectorDatabase(APIObject):
    """
    Metadata for a DataRobot vector database accessible to the user.

    Attributes
    ----------
    id : str
        Vector database ID.
    name : str
        Vector database name.
    size : int
        Size of the vector database assets in bytes.
    use_case_id : str
        Linked use case ID.
    dataset_id : str
        ID of the dataset used for creation.
    embedding_model : str
        Name of the text embedding model.
        Currently supported options are listed in VectorDatabaseEmbeddingModel
        but the values can differ with different platform versions.
    chunking_method : str
        Name of the method to split dataset documents.
        Currently supported options are listed in VectorDatabaseChunkingMethod
        but the values can differ with different platform versions.
    chunk_size : int
        Size of each text chunk in number of tokens.
    chunk_overlap_percentage : int
        Overlap percentage between chunks.
    chunks_count : int
        Total number of text chunks.
    separators : list[string]
        Separators for document splitting.
    creation_date : str
        Date when the database was created.
    creation_user_id : str
        ID of the creating user.
    organization_id : str
        Creating user's organization ID.
    tenant_id : str
        Creating user's tenant ID.
    last_update_date : str
        Last update date for the database.
    execution_status : str
        Database execution status.
        Currently supported options are listed in VectorDatabaseExecutionStatus
        but the values can differ with different platform versions.
    playgrounds_count : int
        Number of using playgrounds.
    dataset_name : str
        Name of the used dataset.
    user_name : str
        Name of the creating user.
    source : str
        Source of the vector database.
        Currently supported options are listed in VectorDatabaseSource
        but the values can differ with different platform versions.
    validation_id : Optional[str]
        ID of custom model vector database validation.
        Only filled for external vector databases.
    error_message : Optional[str]
        Additional information for errored vector database.
    is_separator_regex : bool
        Whether the separators should be treated as regular expressions.
    """

    _path = "api/v2/genai/vectorDatabases"

    _converter = vector_database_trafaret

    def __init__(
        self,
        id: str,
        name: str,
        size: int,
        use_case_id: str,
        dataset_id: Optional[str],
        embedding_model: Optional[str],
        chunking_method: Optional[str],
        chunk_size: int,
        chunk_overlap_percentage: int,
        chunks_count: int,
        separators: List[str],
        creation_date: str,
        creation_user_id: str,
        organization_id: str,
        tenant_id: str,
        last_update_date: str,
        execution_status: str,
        playgrounds_count: int,
        dataset_name: str,
        user_name: str,
        source: str,
        validation_id: Optional[str],
        error_message: Optional[str],
        is_separator_regex: bool,
    ):
        self.id = id
        self.name = name
        self.size = size
        self.use_case_id = use_case_id
        self.dataset_id = dataset_id
        self.embedding_model = embedding_model
        self.chunking_method = chunking_method
        self.chunk_size = chunk_size
        self.chunk_overlap_percentage = chunk_overlap_percentage
        self.chunks_count = chunks_count
        self.separators = separators
        self.creation_date = creation_date
        self.creation_user_id = creation_user_id
        self.organization_id = organization_id
        self.tenant_id = tenant_id
        self.last_update_date = last_update_date
        self.execution_status = execution_status
        self.playgrounds_count = playgrounds_count
        self.dataset_name = dataset_name
        self.user_name = user_name
        self.source = source
        self.validation_id = validation_id
        self.error_message = error_message
        self.is_separator_regex = is_separator_regex

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(id={self.id}, name={self.name}, "
            f"execution_status={self.execution_status})"
        )

    @classmethod
    def create(
        cls,
        dataset_id: str,
        chunking_parameters: ChunkingParameters,
        use_case: Optional[UseCase | str] = None,
        name: Optional[str] = None,
    ) -> VectorDatabase:
        """
        Create a new vector database.

        Parameters
        ----------
        dataset_id : str
            ID of the dataset used for creation.
        chunking_parameters : ChunkingParameters
            Parameters defining how documents are split and embedded.
        use_case : Optional[Union[UseCase, str]], optional
            Use case to link to the created vector database.
        name : str, optional
            Vector database name, by default None
            which leads to the default name 'Vector Database for <dataset name>'.

        Returns
        -------
        vector database : VectorDatabase
            The created vector database with execution status 'new'.
        """
        payload = {
            "name": name,
            "dataset_id": dataset_id,
            "use_case_id": get_use_case_id(use_case, is_required=True),
            "chunking_parameters": chunking_parameters,
        }
        url = f"{cls._client.domain}/{cls._path}/"
        r_data = cls._client.post(url, data=payload)
        return cls.from_server_data(r_data.json())

    @classmethod
    def create_from_custom_model(
        cls,
        name: str,
        use_case: Optional[UseCase | str] = None,
        validation_id: Optional[str] = None,
        prompt_column_name: Optional[str] = None,
        target_column_name: Optional[str] = None,
        deployment_id: Optional[str] = None,
        model_id: Optional[str] = None,
    ) -> VectorDatabase:
        """
        Create a new vector database from validated custom model deployment.

        Parameters
        ----------
        name : str
            Vector database name.
        use_case : Optional[Union[UseCase, str]], optional
            Use case to link to the created vector database.
        validation_id : str, optional
            ID of CustomModelVectorDatabaseValidation for the deployment.
            Alternatively, you can specify ALL the following fields.
        prompt_column_name : str, optional
            The column name the deployed model expect as the input.
        target_column_name : str, optional
            The target name deployed model will output.
        deployment_id : str, optional
            ID of the deployment.
        model_id : str, optional
            ID of the underlying deployment model.
            Can be found from the API as Deployment.model["id"].

        Returns
        -------
        vector database : VectorDatabase
            The created vector database.
        """
        payload: Dict[str, Any] = {
            "use_case_id": get_use_case_id(use_case, is_required=True),
            "name": name,
        }
        if validation_id:
            payload["validation_id"] = validation_id
        else:
            assert all([prompt_column_name, target_column_name, deployment_id, model_id]), (
                "Either validation_id or all of prompt_column_name, target_column_name, "
                "deployment_id, model_id fields should be specified"
            )
            payload.update(
                {
                    "prompt_column_name": prompt_column_name,
                    "target_column_name": target_column_name,
                    "deployment_id": deployment_id,
                    "model_id": model_id,
                }
            )
        url = f"{cls._client.domain}/{cls._path}/fromCustomModelDeployment/"
        r_data = cls._client.post(url, data=payload)
        return cls.from_server_data(r_data.json())

    @classmethod
    def get(cls, vector_database_id: str) -> VectorDatabase:
        """
        Retrieve a single vector database.

        Parameters
        ----------
        vector_database_id : str
            The ID of the vector database you want to retrieve.

        Returns
        -------
        vector database : VectorDatabase
            The requested vector database.
        """
        url = f"{cls._client.domain}/{cls._path}/{vector_database_id}/"
        r_data = cls._client.get(url)
        return cls.from_server_data(r_data.json())

    @classmethod
    def list(
        cls,
        use_case: Optional[UseCaseLike] = None,
        playground: Optional[Union[Playground, str]] = None,
        search: Optional[str] = None,
        sort: Optional[str] = None,
        completed_only: Optional[bool] = None,
    ) -> List[VectorDatabase]:
        """
        List all vector databases associated with a specific use case available to the user.

        Parameters
        ----------
        use_case : Optional[UseCaseLike], optional
            The returned vector databases are filtered to those associated with a specific Use Case
            or Cases if specified or can be inferred from the Context.
            Accepts either the entity or the ID.
        playground : Optional[Union[Playground, str]], optional
            The returned vector databases are filtered to those associated with a specific playground
            if it is specified. Accepts either the entity or the ID.
        search : str, optional
            String for filtering vector databases.
            Vector databases that contain the string in name will be returned.
            If not specified, all vector databases will be returned.
        sort : str, optional
            Property to sort vector databases by.
            Prefix the attribute name with a dash to sort in descending order,
            e.g. sort='-creationDate'.
            Currently supported options are listed in ListVectorDatabasesSortQueryParams
            but the values can differ with different platform versions.
            By default, the sort parameter is None which will result in
            vector databases being returned in order of creation time descending.
        completed_only : bool, optional
            A filter to retrieve only vector databases that have been successfully created.
            By default, all vector databases regardless of execution status are retrieved.

        Returns
        -------
        vectorbases : list[VectorDatabase]
            A list of vector databases available to the user.
        """
        params = {
            "search": search,
            "sort": sort,
            "completed_only": completed_only,
            "playground_id": get_entity_id(playground) if playground else None,
        }

        params = resolve_use_cases(use_cases=use_case, params=params, use_case_key="use_case_id")

        url = f"{cls._client.domain}/{cls._path}/"
        r_data = unpaginate(url, params, cls._client)
        return [cls.from_server_data(data) for data in r_data]

    def update(self, name: str) -> VectorDatabase:
        """
        Update the vector database.

        Parameters
        ----------
        name : str
            The new name for the vector database.

        Returns
        -------
        vector database : VectorDatabase
            The updated vector database.
        """
        payload = {"name": name}
        url = f"{self._client.domain}/{self._path}/{self.id}/"
        r_data = self._client.patch(url, data=payload)
        return self.from_server_data(r_data.json())

    def delete(self) -> None:
        """
        Delete the vector database.
        """
        url = f"{self._client.domain}/{self._path}/{self.id}/"
        self._client.delete(url)

    @classmethod
    def get_supported_embeddings(cls, dataset_id: Optional[str] = None) -> SupportedEmbeddings:
        """Get all supported and the recommended embedding models.

        Parameters
        ----------
        dataset_id : str, optional
            ID of a dataset for which the recommended model is returned
            based on the detected language of that dataset.

        Returns
        -------
        supported_embeddings : SupportedEmbeddings
            The supported embedding models.
        """
        params = {"dataset_id": dataset_id}
        url = f"{cls._client.domain}/{cls._path}/supportedEmbeddings/"
        r_data = cls._client.get(url, params=params)
        return SupportedEmbeddings.from_server_data(r_data.json())

    @classmethod
    def get_supported_text_chunkings(cls) -> SupportedTextChunkings:
        """Get all supported text chunking configurations which includes
        a set of recommended chunking parameters for each supported embedding model.

        Returns
        -------
        supported_text_chunkings : SupportedTextChunkings
            The supported text chunking configurations.
        """
        url = f"{cls._client.domain}/{cls._path}/supportedTextChunkings/"
        r_data = cls._client.get(url)
        return SupportedTextChunkings.from_server_data(r_data.json())

    def download_text_and_embeddings_asset(self, file_path: Optional[str] = None) -> None:
        """Download a parquet file with text chunks and corresponding embeddings created
        by a vector database.

        Parameters
        ----------
        file_path : str, optional
            File path to save the asset. By default, it saves in the current directory
            autogenerated by server name.
        """
        url = f"{self._client.domain}/{self._path}/{self.id}/textAndEmbeddings/"
        response = self._client.get(url, stream=True)
        if not file_path:
            file_path = response.headers["Content-Disposition"].split("=")[-1].strip('"')
        with open(str(file_path), mode="wb") as f:
            for chunk in response.iter_content(chunk_size=65536):
                f.write(chunk)


class CustomModelVectorDatabaseValidation(CustomModelValidation):
    """
    Validation record checking the ability of the deployment to serve
    as a vector database.

    Attributes
    ----------
    prompt_column_name : str
        The column name the deployed model expect as the input.
    target_column_name : str
        The target name deployed model will output.
    deployment_id : str
        ID of the deployment.
    model_id : str
        ID of the underlying deployment model.
        Can be found from the API as Deployment.model["id"].
    validation_status : str
        Can be TESTING, FAILED and PASSED. Only PASSED allowed for use.
    deployment_access_data : dict, optional
        Data that will be used for accessing deployment prediction server.
        Only available for deployments that passed validation. Dict fields:
        - prediction_api_url - URL for deployment prediction server.
        - datarobot_key - first of 2 auth headers for prediction server.
        - authorization_header - second of 2 auth headers for prediction server.
        - input_type - Either JSON or CSV - input type model expects.
        - model_type - Target type of deployed custom model.
    tenant_id : str
        Creating user's tenant ID.
    error_message : Optional[str]
        Additional information for errored validation.
    deployment_name : Optional[str]
        The name of the deployment that is validated.
    user_name : Optional[str]
        The name of the user
    use_case_id : Optional[str]
        The ID of the use case associated with the validation.
    """

    _path = "api/v2/genai/customModelVectorDatabaseValidations"



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

import os
from typing import Any, Dict, List, Optional

import trafaret as t

from datarobot._compat import String
from datarobot.enums import DataDriverListTypes, DataDriverTypes
from datarobot.models.api_object import APIObject


class DataDriver(APIObject):
    """A data driver

    Attributes
    ----------
    id : str
        the id of the driver.
    class_name : str
        the Java class name for the driver.
    canonical_name : str
        the user-friendly name of the driver.
    creator : str
        the id of the user who created the driver.
    base_names : list of str
        a list of the file name(s) of the jar files.
    """

    _path = "externalDataDrivers/"
    _file_upload_path = "externalDataDriverFile/"
    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("class_name", optional=True): t.Or(String, t.Null),
            t.Key("canonical_name"): String(),
            t.Key("creator"): String(),
            t.Key("base_names", optional=True): t.List(String()),
            t.Key("database_driver", optional=True): t.Or(String, t.Null),
            t.Key("type", default="jdbc", optional=True): String(),
            t.Key("version", optional=True): t.Or(String, t.Null),
        }
    ).allow_extra("*")

    def __init__(
        self,
        id: Optional[str] = None,
        creator: Optional[str] = None,
        base_names: Optional[List[str]] = None,
        class_name: Optional[str] = None,
        canonical_name: Optional[str] = None,
        database_driver: Optional[str] = None,
        type: Optional[str] = None,
        version: Optional[str] = None,
    ) -> None:
        self._id = id
        self._creator = creator
        self._base_names = base_names
        self.class_name = class_name
        self.canonical_name = canonical_name
        self.database_driver = database_driver
        self.type = type
        self.version = version

    @classmethod
    def list(cls, typ: Optional[DataDriverListTypes] = None) -> List[DataDriver]:
        """
        Returns list of available drivers.

        Parameters
        ----------
        typ : DataDriverListTypes
            If specified, filters by specified driver type.

        Returns
        -------
        drivers : list of DataDriver instances
            contains a list of available drivers.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> drivers = dr.DataDriver.list()
            >>> drivers
            [DataDriver('mysql'), DataDriver('RedShift'), DataDriver('PostgreSQL')]
        """
        if typ is not None:
            r_data = cls._client.get(cls._path, params={"type": str(typ)}).json()
        else:
            r_data = cls._client.get(cls._path).json()
        return [cls.from_server_data(item) for item in r_data["data"]]

    @classmethod
    def get(cls, driver_id: str) -> DataDriver:
        """
        Gets the driver.

        Parameters
        ----------
        driver_id : str
            the identifier of the driver.

        Returns
        -------
        driver : DataDriver
            the required driver.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> driver = dr.DataDriver.get('5ad08a1889453d0001ea7c5c')
            >>> driver
            DataDriver('PostgreSQL')
        """
        return cls.from_location(f"{cls._path}{driver_id}/")

    @classmethod
    def create(
        cls,
        class_name: str,
        canonical_name: str,
        files: Optional[List[str]] = None,
        typ: Optional[DataDriverTypes] = None,
        database_driver: Optional[str] = None,
    ) -> DataDriver:
        """
        Creates the driver. Only available to admin users.

        Parameters
        ----------
        class_name : str
            the Java class name for the driver. Specify None if typ is DataDriverTypes.DR_DATABASE_V1`.
        canonical_name : str
            the user-friendly name of the driver.
        files : list of str
            a list of the file paths on file system file_path(s) for the driver.
        typ: str
            Optional. Specify the type of the driver. Defaults to `DataDriverTypes.JDBC`, may also be
            `DataDriverTypes.DR_DATABASE_V1`.
        database_driver: str
            Optional. Specify when typ is `DataDriverTypes.DR_DATABASE_V1` to create a native database
            driver. See `DrDatabaseV1Types` enum for some of the types, but that list may not be exhaustive.

        Returns
        -------
        driver : DataDriver
            the created driver.

        Raises
        ------
        ClientError
            raised if user is not granted for `Can manage JDBC database drivers` feature

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> driver = dr.DataDriver.create(
            ...     class_name='org.postgresql.Driver',
            ...     canonical_name='PostgreSQL',
            ...     files=['/tmp/postgresql-42.2.2.jar']
            ... )
            >>> driver
            DataDriver('PostgreSQL')
        """
        payload: Dict[str, Any] = {}
        if typ and typ == DataDriverTypes.DR_DATABASE_V1:
            payload = {
                "type": str(typ),
                "databaseDriver": database_driver,
            }
        else:
            base_names: List[str] = []
            local_jar_urls: List[str] = []

            if files is not None:
                for file_path in files:
                    name = file_path.split(os.sep)[-1]
                    resp = cls._client.build_request_with_file(
                        "POST", cls._file_upload_path, name, file_path=file_path
                    ).json()
                    base_names.append(name)
                    local_jar_urls.append(resp["localUrl"])
            payload = {
                "className": class_name,
                "canonicalName": canonical_name,
                "localJarUrls": local_jar_urls,
                "baseNames": base_names,
            }
        return cls.from_server_data(cls._client.post(cls._path, data=payload).json())

    def update(
        self, class_name: Optional[str] = None, canonical_name: Optional[str] = None
    ) -> None:
        """
        Updates the driver. Only available to admin users.

        Parameters
        ----------
        class_name : str
            the Java class name for the driver.
        canonical_name : str
            the user-friendly name of the driver.

        Raises
        ------
        ClientError
            raised if user is not granted for `Can manage JDBC database drivers` feature

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> driver = dr.DataDriver.get('5ad08a1889453d0001ea7c5c')
            >>> driver.canonical_name
            'PostgreSQL'
            >>> driver.update(canonical_name='postgres')
            >>> driver.canonical_name
            'postgres'
        """
        payload = {
            "className": class_name or self.class_name,
            "canonicalName": canonical_name or self.canonical_name,
        }
        r_data = self._client.patch(f"{self._path}{self.id}/", data=payload).json()
        self.class_name = r_data["className"]
        self.canonical_name = r_data["canonicalName"]

    def delete(self) -> None:
        """
        Removes the driver. Only available to admin users.

        Raises
        ------
        ClientError
            raised if user is not granted for `Can manage JDBC database drivers` feature
        """
        self._client.delete(f"{self._path}{self.id}/")

    @property
    def id(self) -> Optional[str]:
        return self._id

    @property
    def creator(self) -> Optional[str]:
        return self._creator

    @property
    def base_names(self) -> Optional[List[str]]:
        return self._base_names

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self.canonical_name or self.id}')"



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Any, cast, Dict, List

import trafaret as t

from datarobot.insights.base import BaseInsight


class ShapPreview(BaseInsight):
    """Class for SHAP Preview calculations. Use the standard methods of BaseInsight to compute and
    retrieve SHAP matrices:
    * compute: submit a request to compute a SHAP preview, and return immediately
    * create: submit a request to compute a SHAP preview, and wait for it to finish
    * list: retrieve all ShapPreview results for a model, possibly on multiple datasets or data slices.
    """

    SHAP_PREVIEW_VALUE = t.Dict(
        {
            t.Key("feature_rank"): t.Int(),
            t.Key("feature_name"): t.String(),
            t.Key("feature_value"): t.String(),
            t.Key("shap_value"): t.Float(),
        }
    ).ignore_extra("*")

    SHAP_PREVIEW_ROW = t.Dict(
        {
            t.Key("row_index"): t.Int(),
            t.Key("total_preview_features"): t.Int(),
            t.Key("prediction_value"): t.Float(),
            t.Key("preview_values"): t.List(SHAP_PREVIEW_VALUE),
        }
    )

    INSIGHT_NAME = "shapPreview"
    INSIGHT_DATA = {
        t.Key("previews_count"): t.Int(),
        t.Key("previews"): t.List(SHAP_PREVIEW_ROW),
    }

    @property
    def previews(self) -> List[Dict[str, Any]]:
        """SHAP preview values

        Returns
        -------
        preview : List[Dict[str, Any]]
            A list of the ShapPreview values for each row
        """
        return cast(List[Dict[str, Any]], self.data["previews"])

    @property
    def previews_count(self) -> int:
        """Number of shap preview rows

        Returns
        -------
        int
        """
        return cast(int, self.data["previews_count"])



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Any, cast, Dict, Iterable, List, Optional, Union

import trafaret as t
from typing_extensions import Self

from datarobot.enums import DEFAULT_MAX_WAIT, ENTITY_TYPES, INSIGHTS_SOURCES
from datarobot.mixins.browser_mixin import BrowserMixin
from datarobot.models import StatusCheckJob
from datarobot.models.api_object import APIObject
from datarobot.utils import pagination


class BaseInsight(APIObject, BrowserMixin):
    """Base Insight class for modern insights

    This class serves as a template for modern insights created using the Root Insights framework.
    It provides most necessary functions for easily implementing classes that wrap specific insights.

    """

    # To use this class it should be inherited and the `INSIGHT_NAME` and `INSIGHT_DATA` attributes
    # should be replaced with the correct values.
    #
    # Example:
    # You have an insight that has the endpoint `insights/myInsight` and the insight's calculator
    # produces a response that contains a boolean field called `insight_worked` and a list of
    # string values called `insight_values`.
    #
    # To make things easier for the user you may want to add some easily accessible properties
    # to allow them to quickly get the values of the insight.
    #
    # Example insight result:
    #
    # .. code-block:: text
    #
    #         {
    #             "insight_worked": true,
    #             "insight_values": ["hello", "world"]
    #         }
    #
    # You can simply create a wrapper for this new insight using this framework:
    #
    # .. code-block:: python
    #
    #         class MyInsight(BaseInsight):
    #             INSIGHT_NAME = "myInsight"
    #             INSIGHT_DATA = {
    #                 t.Key("insight_worked"): t.Bool(),
    #                 t.Key("insight_values"): t.List(t.String()),
    #             }
    #
    #             @property
    #             def values() -> List[str]:
    #                 return cast(List[str], self.data["insight_values"])
    #
    # All of the computation and insight retrieval handling will be accounted for by the base insight.
    # This class can also be overridden for more complex queries by adjusting the functionality of the
    # `_get_payload` function to provide additional required options, but most insights won't need
    # this capability.

    INSIGHT_NAME: str = "base"  # Defined by subclasses to identify the formal name for endpoints
    INSIGHT_DATA = t.Any()  # Defined by subclasses for specific insights

    _converter = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("entity_id"): t.String(),
            t.Key("project_id"): t.String(),
            t.Key("source"): t.String(),
            t.Key("data_slice_id", optional=True): t.Or(t.String(), t.Null()),
            t.Key("external_dataset_id", optional=True): t.Or(t.String(), t.Null()),
            t.Key("data"): INSIGHT_DATA,
        }
    ).ignore_extra("*")

    def __init__(
        self,
        id: str,
        entity_id: str,
        project_id: str,
        source: str,
        data: Any,
        data_slice_id: Optional[str] = None,
        external_dataset_id: Optional[str] = None,
    ):
        """Base Insight class

        Parameters
        ----------
        id: str
            ID of the insight.
        entity_id: str
            ID of the entity associated with the insight.
        project_id: str
            ID of the project associated with the insight.
        source: str
            Source type of the insight.
        data: Dict
            Results of the computed insight.
        data_slice_id: Optional[str]
            Data slice ID associated with the insight.
        external_dataset_id: Optional[str]
            External dataset ID associated with the insight.
        """
        self.id = id
        self.entity_id = entity_id
        self.project_id = project_id
        self.source = source
        self.data_slice_id = data_slice_id
        self.external_dataset_id = external_dataset_id

        self.data = data

    @classmethod
    def _compute_url(cls) -> str:
        """Construct a compute url"""
        return f"insights/{cls.INSIGHT_NAME}/"

    @classmethod
    def _retrieve_url(cls) -> str:
        """Construct a retrieve url"""
        return f"insights/{cls.INSIGHT_NAME}" + "/models/{}/"

    def get_uri(self) -> str:
        """This should define the URI to their browser based interactions"""
        raise NotImplementedError

    @classmethod
    # type: ignore[override]
    def from_server_data(
        cls,
        data: Dict[str, Any],
        keep_attrs: Optional[Iterable[str]] = None,
    ) -> Self:
        """Override from_server_data to handle paginated responses"""
        if all(attrib in data for attrib in ["count", "next", "previous"]):
            # If true, the data response is a paginated response from a compute command and the data
            # entity should be unwrapped. There should only be one response.
            data = data["data"][0]

        return super().from_server_data(data=data, keep_attrs=keep_attrs)

    @classmethod
    def _get_payload(
        cls,
        entity_id: str,
        source: str = INSIGHTS_SOURCES.VALIDATION,
        data_slice_id: Optional[str] = None,
        external_dataset_id: Optional[str] = None,
        entity_type: Optional[ENTITY_TYPES] = None,
        **kwargs: Any,
    ) -> Dict[str, Union[str, int]]:
        """Construct a payload for a compute request. May be overridden by insight subclasses to
        accept additional parameters.
        """
        _ = kwargs
        payload = {
            "entityId": entity_id,
            "source": source,
        }
        if data_slice_id is not None:
            payload["dataSliceId"] = data_slice_id
        if external_dataset_id is not None:
            payload["externalDatasetId"] = external_dataset_id
        if entity_type is not None:
            payload["entityType"] = entity_type

        return cast(Dict[str, Union[str, int]], payload)

    @classmethod
    def compute(
        cls,
        entity_id: str,
        source: str = INSIGHTS_SOURCES.VALIDATION,
        data_slice_id: Optional[str] = None,
        external_dataset_id: Optional[str] = None,
        entity_type: Optional[ENTITY_TYPES] = ENTITY_TYPES.DATAROBOT_MODEL,
        **kwargs: Any,
    ) -> StatusCheckJob:
        """Submit an insight compute request. You can use `create` if you want to
        wait synchronously for the completion of the job. May be overridden by insight subclasses to
        accept additional parameters.

        Parameters
        ----------
        entity_id: str
            ID of the entity to compute the insight.
        source: str
            Source type to use when computing the insight.
        data_slice_id: Optional[str]
            Data slice ID to use when computing the insight.
        external_dataset_id: Optional[str]
            External dataset ID to use when computing the insight.
        entity_type: Optional[ENTITY_TYPES]
            Type of the entity associated with the insight. Select one of the ENTITY_TYPE enum
            values, or accept the default, "datarobotModel".

        Returns
        -------
        StatusCheckJob
            Status check job entity for the asynchronous insight calculation.
        """
        payload = cls._get_payload(
            entity_id=entity_id,
            source=source,
            data_slice_id=data_slice_id,
            external_dataset_id=external_dataset_id,
            entity_type=entity_type,
            **kwargs,
        )
        response = cls._client.post(cls._compute_url(), data=payload)

        return StatusCheckJob.from_response(response, cls)

    @classmethod
    def create(
        cls,
        entity_id: str,
        source: str = INSIGHTS_SOURCES.VALIDATION,
        data_slice_id: Optional[str] = None,
        external_dataset_id: Optional[str] = None,
        entity_type: Optional[ENTITY_TYPES] = ENTITY_TYPES.DATAROBOT_MODEL,
        max_wait: Optional[int] = DEFAULT_MAX_WAIT,
        **kwargs: Any,
    ) -> Self:
        """Create an insight and wait for completion. May be overridden by insight subclasses to
        accept additional parameters.

        Parameters
        ----------
        entity_id: str
            ID of the entity to compute the insight.
        source: str
            Source type to use when computing the insight.
        data_slice_id: Optional[str]
            Data slice ID to use when computing the insight.
        external_dataset_id: Optional[str]
            External dataset ID to use when computing the insight.
        entity_type: Optional[ENTITY_TYPES]
            Type of the entity associated with the insight. Select one of the ENTITY_TYPE enum
            values, or accept the default, "datarobotModel".
        max_wait: int
            Number of seconds to wait for the result.

        Returns
        -------
        Self
            Entity of the newly or already computed insights.
        """
        status_check_job = cls.compute(
            entity_id=entity_id,
            source=source,
            data_slice_id=data_slice_id,
            external_dataset_id=external_dataset_id,
            entity_type=entity_type,
            **kwargs,
        )
        return cast(Self, status_check_job.get_result_when_complete(max_wait=cast(int, max_wait)))

    @classmethod
    def list(cls, entity_id: str) -> List[Self]:
        """List all generated insights.

        Parameters
        ----------
        entity_id: str
            ID of the entity to list all generated insights.

        Returns
        -------
        List[Any]
            List of newly or already computed insights.
        """
        url = cls._retrieve_url().format(entity_id)
        return [cls.from_server_data(x) for x in pagination.unpaginate(url, {}, cls._client)]



# flake8: noqa
# because the unused imports are on purpose

from .base import BaseInsight
from .shap_impact import ShapImpact
from .shap_matrix import ShapMatrix
from .shap_preview import ShapPreview



#
# Copyright 2021-2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.

# flake8: noqa

from .job import Job, JobFileItem
from .job_run import JobRun, JobRunStatus



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

import enum
import inspect
from typing import Any, Dict, List, Optional, TYPE_CHECKING, Union

from mypy_extensions import TypedDict
import trafaret as t

from datarobot._compat import Int, String
from datarobot.enums import enum_to_list, INSIGHTS_SOURCES
from datarobot.models.api_object import APIObject
from datarobot.models.status_check_job import StatusCheckJob
from datarobot.utils.pagination import unpaginate

if TYPE_CHECKING:
    from datarobot.models.model import Model
    from datarobot.models.project import Project


class DataSliceFiltersType(TypedDict):
    operand: str
    operator: str
    values: List[Union[str, int, float]]


class DataSliceSizeMessageType(TypedDict):
    level: str
    description: str
    additional_info: str


class DataSlicesOperators(str, enum.Enum):
    EQUAL = "eq"
    CONTAINS = "in"
    LESS_THAN = "<"
    GREATER_THAN = ">"
    BETWEEN = "between"
    NOT_BETWEEN = "notBetween"


class DataSliceSizeInfo(APIObject):
    """
    Definition of a data slice applied to a source

    Attributes
    ----------
    data_slice_id : str
        ID of the data slice
    project_id : str
        ID of the project
    source : str
        Data source used to calculate the number of rows (slice size) after applying the data slice's filters
    model_id : str, optional
        ID of the model, required when source (subset) is 'training'
    slice_size : int
        Number of rows in the data slice for a given source
    messages : list[DataSliceSizeMessageType]
        List of user-relevant messages related to a data slice
    """

    _converter = t.Dict(
        {
            t.Key("data_slice_id", optional=True): String,
            t.Key("project_id", optional=True): String,
            t.Key("source", optional=True): t.Enum(*enum_to_list(INSIGHTS_SOURCES)),
            t.Key("model_id", optional=True): String,
            t.Key("slice_size", optional=True): Int,
            t.Key("messages", optional=True): t.Or(
                t.Null,
                t.List(
                    t.Dict(
                        {
                            t.Key("level"): String,
                            t.Key("description"): String,
                            t.Key("additional_info"): String,
                        }
                    ).ignore_extra("*"),
                    min_length=1,
                    max_length=3,
                ),
            ),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        data_slice_id: Optional[str] = None,
        project_id: Optional[str] = None,
        source: Optional[str] = None,
        slice_size: Optional[int] = None,
        messages: Optional[List[DataSliceSizeMessageType]] = None,
        model_id: Optional[str] = None,
    ) -> None:
        self.data_slice_id = data_slice_id
        self.project_id = project_id
        self.source = source
        self.slice_size = slice_size
        self.messages = messages
        self.model_id = model_id

    def __repr__(self) -> str:
        slice_size_info_dict = self.to_dict()
        data = ", ".join(f"{key}={value}" for key, value in slice_size_info_dict.items())
        return f"{self.__class__.__name__}({data})"

    def to_dict(self) -> Dict[str, Any]:
        return {
            "data_slice_id": self.data_slice_id,
            "project_id": self.project_id,
            "source": self.source,
            "slice_size": self.slice_size,
            "messages": self.messages,
            "model_id": self.model_id,
        }


class DataSlice(APIObject):
    """
    Definition of a data slice

    Attributes
    ----------
    id : str
        ID of the data slice.
    name : str
        Name of the data slice definition.
    filters : list[DataSliceFiltersType]
        List of filters (dict) with params:
            - operand : str
                Name of the feature to use in the filter.
            - operator : str
                Operator to use in the filter: 'eq', 'in', '<', or '>'.
            - values : Union[str, int, float]
                Values to use from the feature.
    project_id : str
        ID of the project that the model is part of.
    """

    _base_data_slices_path_template = "dataSlices"
    _data_slices_sizes_path_template = (
        _base_data_slices_path_template + "/{data_slice_id}/sliceSizes"
    )

    FilterDataSlicesDefinition = t.Dict(
        {
            t.Key("operand"): String,
            t.Key("operator"): t.Enum(*enum_to_list(DataSlicesOperators)),
            t.Key("values"): t.List(t.Or(String, Int, t.Float), min_length=1, max_length=1000),
        }
    )

    _converter = t.Dict(
        {
            t.Key("id", optional=True): String,
            t.Key("name", optional=True): String,
            t.Key("filters", optional=True): t.List(
                FilterDataSlicesDefinition, min_length=1, max_length=3
            ),
            t.Key("project_id", optional=True): String,
        }
    ).ignore_extra("*")

    def __init__(
        self,
        id: Optional[str] = None,
        name: Optional[str] = None,
        filters: Optional[List[DataSliceFiltersType]] = None,
        project_id: Optional[str] = None,
    ) -> None:
        self.id = id
        self.name = name
        self.filters = filters
        self.project_id = project_id

    def __repr__(self) -> str:
        slice_dict = self.to_dict()
        data = ", ".join(f"{key}={value}" for key, value in slice_dict.items())
        return f"{self.__class__.__name__}({data})"

    def to_dict(self) -> Dict[str, Any]:
        return {
            attr[0]: attr[1]
            for attr in inspect.getmembers(self)
            if (
                not attr[0].startswith("_")
                and not inspect.ismethod(attr[1])
                and not isinstance(attr[1], type(self.FilterDataSlicesDefinition))
            )
        }

    @classmethod
    def list(
        cls,
        project: Union[str, Project],
        offset: Optional[int] = 0,
        limit: Optional[int] = 100,
    ) -> List[DataSlice]:
        """
        List the data slices in the same project

        Parameters
        ----------
        project : Union[str, Project]
            ID of the project or Project object from which to list data slices.
        offset : int, optional
            Number of items to skip.
        limit : int, optional
            Number of items to return.

        Returns
        -------
        data_slices : list[DataSlice]

        Examples
        --------

        .. code-block:: python

            >>> import datarobot as dr
            >>> ...  # set up your Client
            >>> data_slices = dr.DataSlice.list("646d0ea0cd8eb2355a68b0e5")
            >>> data_slices
            [DataSlice(...), DataSlice(...), ...]
        """
        project_id = project if isinstance(project, str) or project is None else project.id
        url = f"projects/{project_id}/{cls._base_data_slices_path_template}"
        query_params: Dict[str, Union[Optional[int], Optional[str]]] = {
            "offset": offset,
            "limit": limit,
        }
        data_slices = [
            cls.from_server_data(item) for item in unpaginate(url, query_params, cls._client)
        ]

        return data_slices

    @classmethod
    def create(
        cls, name: str, filters: List[DataSliceFiltersType], project: Union[str, Project]
    ) -> DataSlice:
        """
        Creates a data slice in the project with the given name and filters

        Parameters
        ----------
        name : str
            Name of the data slice definition.
        filters : list[DataSliceFiltersType]
            List of filters (dict) with params:
                - operand : str
                    Name of the feature to use in filter.
                - operator : str
                    Operator to use: 'eq', 'in', '<', or '>'.
                - values : Union[str, int, float]
                    Values to use from the feature.
        project : Union[str, Project]
            Project ID or Project object from which to list data slices.

        Returns
        -------
        data_slice : DataSlice
            The data slice object created

        Examples
        --------

        .. code-block:: python

            >>> import datarobot as dr
            >>> ...  # set up your Client and retrieve a project
            >>> data_slice = dr.DataSlice.create(
            >>> ...    name='yes',
            >>> ...    filters=[{'operand': 'binary_target', 'operator': 'eq', 'values': ['Yes']}],
            >>> ...    project=project,
            >>> ...  )
            >>> data_slice
            DataSlice(
                filters=[{'operand': 'binary_target', 'operator': 'eq', 'values': ['Yes']}],
                id=646d1296bd0c543d88923c9d,
                name=yes,
                project_id=646d0ea0cd8eb2355a68b0e5
            )
        """
        project_id = project if isinstance(project, str) or project is None else project.id
        data = {"name": name, "filters": filters, "project_id": project_id}
        response = cls._client.post(cls._base_data_slices_path_template, data=data)
        response_json = response.json()
        data_slice = DataSlice.from_server_data(response_json)
        return data_slice

    def delete(self) -> None:
        """
        Deletes the data slice from storage

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> data_slice = dr.DataSlice.get('5a8ac9ab07a57a0001be501f')
            >>> data_slice.delete()

        .. code-block:: python

            >>> import datarobot as dr
            >>> ... # get project or project_id
            >>> data_slices = dr.DataSlice.list(project)  # project object or project_id
            >>> data_slice = data_slices[0]  # choose a data slice from the list
            >>> data_slice.delete()
        """
        url = f"{self._base_data_slices_path_template}/{self.id}"
        self._client.delete(url)

    def request_size(
        self, source: INSIGHTS_SOURCES, model: Optional[Union[str, Model]] = None
    ) -> StatusCheckJob:
        """
        Submits a request to validate the data slice's filters and
        calculate the data slice's number of rows on a given source

        Parameters
        ----------
        source : INSIGHTS_SOURCES
            Subset of data (partition or "source") on which to apply the data slice
            for estimating available rows.
        model : Optional[Union[str, Model]]
            Model object or ID of the model. It is only required when source is "training".

        Returns
        -------
        status_check_job : StatusCheckJob
            Object contains all needed logic for a periodical status check of an async job.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> ... # get project or project_id
            >>> data_slices = dr.DataSlice.list(project)  # project object or project_id
            >>> data_slice = data_slices[0]  # choose a data slice from the list
            >>> status_check_job = data_slice.request_size("validation")

        Model is required when source is 'training'

        .. code-block:: python

            >>> import datarobot as dr
            >>> ... # get project or project_id
            >>> data_slices = dr.DataSlice.list(project)  # project object or project_id
            >>> data_slice = data_slices[0]  # choose a data slice from the list
            >>> status_check_job = data_slice.request_size("training", model)
        """
        model_id = model if isinstance(model, str) or model is None else model.id
        route = self._data_slices_sizes_path_template.format(data_slice_id=self.id)
        data = {"source": source, "project_id": self.project_id, "model_id": model_id}
        response = self._client.post(route, data=data)
        return StatusCheckJob.from_response(response, DataSliceSizeInfo)

    def get_size_info(
        self, source: INSIGHTS_SOURCES, model: Optional[Union[str, Model]] = None
    ) -> DataSliceSizeInfo:
        """
        Get information about the data slice applied to a source

        Parameters
        ----------
        source : INSIGHTS_SOURCES
            Source (partition or subset) to which the data slice was applied
        model : Optional[Union[str, Model]]
            ID for the model whose training data was sliced with this data slice.
            Required when the source is "training", and not used for other sources.

        Returns
        -------
        slice_size_info : DataSliceSizeInfo
            Information of the data slice applied to a source

        Examples
        --------

        .. code-block:: python

            >>> import datarobot as dr
            >>> ...  # set up your Client
            >>> data_slices = dr.DataSlice.list("646d0ea0cd8eb2355a68b0e5")
            >>> data_slice = slices[0]  # can be any slice in the list
            >>> data_slice_size_info = data_slice.get_size_info("validation")
            >>> data_slice_size_info
            DataSliceSizeInfo(
                data_slice_id=6493a1776ea78e6644382535,
                messages=[
                    {
                        'level': 'WARNING',
                        'description': 'Low Observation Count',
                        'additional_info': 'Insufficient number of observations to compute some insights.'
                    }
                ],
                model_id=None,
                project_id=646d0ea0cd8eb2355a68b0e5,
                slice_size=1,
                source=validation,
            )
            >>> data_slice_size_info.to_dict()
            {
                'data_slice_id': '6493a1776ea78e6644382535',
                'messages': [
                    {
                        'level': 'WARNING',
                        'description': 'Low Observation Count',
                        'additional_info': 'Insufficient number of observations to compute some insights.'
                    }
                ],
                'model_id': None,
                'project_id': '646d0ea0cd8eb2355a68b0e5',
                'slice_size': 1,
                'source': 'validation',
            }

        .. code-block:: python

            >>> import datarobot as dr
            >>> ...  # set up your Client
            >>> data_slice = dr.DataSlice.get("6493a1776ea78e6644382535")
            >>> data_slice_size_info = data_slice.get_size_info("validation")

        When using source='training', the model param is required.

        .. code-block:: python

            >>> import datarobot as dr
            >>> ...  # set up your Client
            >>> model = dr.Model.get(project_id, model_id)
            >>> data_slice = dr.DataSlice.get("6493a1776ea78e6644382535")
            >>> data_slice_size_info = data_slice.get_size_info("training", model)

        .. code-block:: python

            >>> import datarobot as dr
            >>> ...  # set up your Client
            >>> data_slice = dr.DataSlice.get("6493a1776ea78e6644382535")
            >>> data_slice_size_info = data_slice.get_size_info("training", model_id)
        """
        model_id = model if isinstance(model, str) or model is None else model.id
        query_params = {"source": source, "projectId": self.project_id}
        if model_id:
            query_params["modelId"] = model_id
        url = self._data_slices_sizes_path_template.format(data_slice_id=self.id)
        response = self._client.get(url, params=query_params)
        if response.status_code == 204:
            # object with no values
            return DataSliceSizeInfo()
        response_json = response.json()
        slice_size_info = DataSliceSizeInfo.from_server_data(response_json)
        return slice_size_info

    @classmethod
    def get(cls, data_slice_id: str) -> DataSlice:
        """
        Retrieve a specific data slice.

        Parameters
        ----------
        data_slice_id : str
            The identifier of the data slice to retrieve.

        Returns
        -------
        data_slice: DataSlice
            The required data slice.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> dr.DataSlice.get('648b232b9da812a6aaa0b7a9')
            DataSlice(filters=[{'operand': 'binary_target', 'operator': 'eq', 'values': ['Yes']}],
                      id=648b232b9da812a6aaa0b7a9,
                      name=test,
                      project_id=644bc575572480b565ca42cd
                      )
        """
        url = f"{cls._base_data_slices_path_template}/{data_slice_id}"
        return cls.from_location(url)



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import List, Optional, Union

import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.models.genai.llm_blueprint import LLMBlueprint
from datarobot.utils.pagination import unpaginate


def get_entity_id(entity: Union[Chat, LLMBlueprint, str]) -> str:
    """
    Get the entity ID from the entity parameter.

    Parameters
    ----------
    entity : APIObject or str
        Specifies either the entity ID or the entity.

    Returns
    -------
    id : str
        The entity ID.
    """
    return entity if isinstance(entity, str) else entity.id


chat_trafaret = t.Dict(
    {
        t.Key("id"): t.String,
        t.Key("name"): t.String,
        t.Key("llm_blueprint_id"): t.String,
        t.Key("is_frozen"): t.Bool,
        t.Key("creation_date"): t.String,
        t.Key("creation_user_id"): t.String,
    }
).ignore_extra("*")


class Chat(APIObject):
    """
    Metadata for a DataRobot GenAI chat.

    Attributes
    ----------
    id : str
        The chat ID.
    name : str
        The chat name.
    llm_blueprint_id : str
        The ID of the LLM blueprint associated with the chat.
    is_frozen : bool
        Checks whether the chat is frozen. Prompts cannot be submitted to frozen chats.
    creation_date : str
        The date when the chat was created.
    creation_user_id : str
        The ID of the creating user.
    """

    _path = "api/v2/genai/chats"

    _converter = chat_trafaret

    def __init__(
        self,
        id: str,
        name: str,
        llm_blueprint_id: str,
        is_frozen: bool,
        creation_date: str,
        creation_user_id: str,
    ):
        self.id = id
        self.name = name
        self.llm_blueprint_id = llm_blueprint_id
        self.is_frozen = is_frozen
        self.creation_date = creation_date
        self.creation_user_id = creation_user_id

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id})"

    @classmethod
    def create(
        cls,
        name: str,
        llm_blueprint: Union[LLMBlueprint, str],
    ) -> Chat:
        """
        Creates a new chat.

        Parameters
        ----------
        name : str
            The chat name.
        llm_blueprint : LLMBlueprint or str
            The LLM blueprint associated with the created chat, either LLM blueprint or ID.

        Returns
        -------
        chat : Chat
            The created chat.
        """
        payload = {
            "llm_blueprint_id": get_entity_id(llm_blueprint),
            "name": name,
        }

        url = f"{cls._client.domain}/{cls._path}/"
        r_data = cls._client.post(url, data=payload)
        return cls.from_server_data(r_data.json())

    @classmethod
    def get(cls, chat: Union[Chat, str]) -> Chat:
        """
        Retrieve a single chat.

        Parameters
        ----------
        chat : Chat or str
            The chat you want to retrieve. Accepts chat or chat ID.

        Returns
        -------
        chat : Chat
            The requested chat.
        """
        url = f"{cls._client.domain}/{cls._path}/{get_entity_id(chat)}/"
        r_data = cls._client.get(url)
        return cls.from_server_data(r_data.json())

    @classmethod
    def list(
        cls,
        llm_blueprint: Optional[Union[LLMBlueprint, str]] = None,
        sort: Optional[str] = None,
    ) -> List[Chat]:
        """
        List all chats available to the user. If the LLM blueprint is specified,
        results are restricted to only those chats associated with the LLM blueprint.

        Parameters
        ----------
        llm_blueprint : Optional[Union[LLMBlueprint, str]], optional
            Returns only those chats associated with a particular LLM blueprint,
            specified by either the entity or the ID.
        sort : str, optional
            The property to sort chats by. Prefix the attribute name with a dash ( - )
            to sort responses in descending order, (for example, '-name').
            Supported options are listed in ListChatsSortQueryParams,
            but the values can differ depending on platform version.
            The default sort parameter is None, which results in
            chats returning in order of creation time, descending.

        Returns
        -------
        chats : list[Chat]
            Returns a list of chats.
        """
        params = {
            "llm_blueprint_id": get_entity_id(llm_blueprint) if llm_blueprint else None,
            "sort": sort,
        }
        url = f"{cls._client.domain}/{cls._path}/"
        r_data = unpaginate(url, params, cls._client)
        return [cls.from_server_data(data) for data in r_data]

    def delete(self) -> None:
        """
        Delete the single chat.
        """
        url = f"{self._client.domain}/{self._path}/{self.id}/"
        self._client.delete(url)

    def update(self, name: str) -> Chat:
        """
        Update the chat.

        Parameters
        ----------
        name : str
            The new name for the chat.

        Returns
        -------
        chat : Chat
            The updated chat.
        """
        payload = {"name": name}
        url = f"{self._client.domain}/{self._path}/{self.id}/"
        r_data = self._client.patch(url, data=payload)
        return self.from_server_data(r_data.json())



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import csv
from io import StringIO
import os
from typing import Any, Dict, List

import pandas as pd

from .. import errors


def dataframe_to_buffer(df: pd.DataFrame) -> StringIO:
    """Convert a dataframe to a serialized form in a buffer

    Parameters
    ----------
    df : pandas.DataFrame
        The data to serialize

    Returns
    -------
    buff : StringIO()
        The data. The descriptor will be reset before being returned (seek(0))
    """
    buff = StringIO()
    df.to_csv(buff, encoding="utf-8", index=False, quoting=csv.QUOTE_ALL)
    buff.seek(0)
    return buff


def list_of_records_to_buffer(list_of_records: List[Dict[str, Any]]) -> StringIO:
    """

    Parameters
    ----------
    list_of_records: list(dict)
        A list of records to serialize. Each row should be a dict with the same keys as the first
        row.

    Returns
    -------
    buffer: StringIO
        The data. The descriptor will be reset before being returned (seek(0))
    """
    buff = StringIO()
    headers = list(list_of_records[0])
    csv_writer = csv.DictWriter(buff, fieldnames=headers)
    csv_writer.writeheader()
    csv_writer.writerows(list_of_records)
    buff.seek(0)
    return buff


def is_urlsource(sourcedata: Any) -> bool:
    """Whether sourcedata is of url kind"""
    return isinstance(sourcedata, str) and (
        sourcedata.startswith("http")
        or sourcedata.startswith("file:")
        or sourcedata.startswith("s3:")
    )


def recognize_sourcedata(sourcedata: Any, default_fname: str) -> Dict[str, Any]:
    """Given a sourcedata figure out if it is a filepath, dataframe, or
    filehandle, and then return the correct kwargs for the upload process
    """
    if isinstance(sourcedata, pd.DataFrame):
        buff = dataframe_to_buffer(sourcedata)
        return {"filelike": buff, "fname": default_fname}
    elif hasattr(sourcedata, "read") and hasattr(sourcedata, "seek"):
        return {"filelike": sourcedata, "fname": default_fname}
    elif isinstance(sourcedata, str) and os.path.isfile(sourcedata):
        return {"file_path": sourcedata, "fname": os.path.split(sourcedata)[1]}
    elif isinstance(sourcedata, bytes) and not is_urlsource(sourcedata):
        assert_modelable(sourcedata)
        return {"content": sourcedata, "fname": default_fname}
    else:
        # Checking if string representation of an object can be used as a file path.
        # The main reason for this is supporting Python's pathlib.Path without having extra
        # dependency for python lower than 3.4.
        try:
            if os.path.isfile(str(sourcedata)):
                return {"file_path": str(sourcedata), "fname": os.path.split(str(sourcedata))[1]}
        except Exception:
            pass

    err_msg = (
        "sourcedata parameter not understood. Use pandas "
        "DataFrame, file object or string that is either a "
        "path to file or raw file content to specify data "
        "to upload"
    )
    raise errors.InputNotUnderstoodError(err_msg)


def assert_modelable(sourcedata: bytes) -> None:
    """
    Uses a heuristic to assert that the given argument is not a
    filepath.

    Some users have mistyped filepaths before, which
    the function `recognize_sourcedata` interpreted as being
    the actual data to use for modeling. This would lead to other
    problems later on.

    Parameters
    ----------
    sourcedata : bytes
        The data which we are trying to assert is not a mistyped
        file path.

    Raises
    ------
    InputNotUnderstoodError
        If this does look like a filepath (it should not)
    """
    if len(sourcedata) < 500 and len(sourcedata.splitlines()) == 1:
        raise errors.InputNotUnderstoodError(
            "The source appears to contain only a single line of data. "
            "Did you mistype a file path?"
        )



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, List, Optional

import trafaret as t

from datarobot.enums import DEFAULT_MAX_WAIT
from datarobot.models.api_object import APIObject
from datarobot.utils import get_id_from_location
from datarobot.utils.waiters import wait_for_async_resolution


class Challenger(APIObject):
    """A challenger is an alternative model being compared to the model currently deployed

    Attributes
    ----------
    id : str
        The ID of the challenger.
    deployment_id : str
        The ID of the deployment.
    name : str
        The name of the challenger.
    model : dict
        The model of the challenger.
    model_package : dict
        The model package of the challenger.
    prediction_environment : dict
        The prediction environment of the challenger.
    """

    _path = "deployments/{}/challengers/"

    _converter = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("name"): t.String(),
            t.Key("model"): t.Dict().allow_extra("*"),
            t.Key("model_package"): t.Dict().allow_extra("*"),
            t.Key("prediction_environment"): t.Dict().allow_extra("*"),
        }
    )

    def __init__(
        self,
        id: str,
        deployment_id: Optional[str] = None,
        name: Optional[str] = None,
        model: Optional[Dict[str, Any]] = None,
        model_package: Optional[Dict[str, Any]] = None,
        prediction_environment: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.id = id
        self.deployment_id = deployment_id
        self.name = name
        self.model = model
        self.model_package = model_package
        self.prediction_environment = prediction_environment

    def __repr__(self) -> str:
        return f"Challenger({self.name})"

    @classmethod
    def create(
        cls,
        deployment_id: str,
        model_package_id: str,
        prediction_environment_id: str,
        name: str,
        max_wait: int = DEFAULT_MAX_WAIT,
    ) -> Challenger:
        """Create a challenger for a deployment

        Parameters
        ----------
        deployment_id : str
            The ID of the deployment
        model_package_id : str
            The model package id of the challenger model
        prediction_environment_id : str
            The prediction environment id of the challenger model
        name : str
            The name of the challenger model
        max_wait : int, optional
            The amount of seconds to wait for successful resolution of a challenger creation job.

        Examples
        --------
        .. code-block:: python

            from datarobot import Challenger
            challenger = Challenger.create(
                deployment_id="5c939e08962d741e34f609f0",
                name="Elastic-Net Classifier",
                model_package_id="5c0a969859b00004ba52e41b",
                prediction_environment_id="60b012436635fc00909df555"
            )
        """
        payload = {
            "modelPackageId": model_package_id,
            "name": name,
            "predictionEnvironmentId": prediction_environment_id,
        }
        path = cls._path.format(deployment_id)
        response = cls._client.post(path, data=payload)
        challenger_loc = wait_for_async_resolution(
            cls._client, response.headers["Location"], max_wait
        )
        return cls.get(deployment_id, get_id_from_location(challenger_loc))

    @classmethod
    def get(cls, deployment_id: str, challenger_id: str) -> Challenger:
        """Get a challenger for a deployment

        Parameters
        ----------
        deployment_id : str
            The ID of the deployment
        challenger_id : str
            The ID of the challenger

        Returns
        -------
        Challenger
            The challenger object

        Examples
        --------
        .. code-block:: python

            from datarobot import Challenger
            challenger = Challenger.get(
                deployment_id="5c939e08962d741e34f609f0",
                challenger_id="5c939e08962d741e34f609f0"
            )

            challenger.id
            >>>'5c939e08962d741e34f609f0'
            challenger.model_package['name']
            >>> 'Elastic-Net Classifier'
        """
        path = cls._path.format(deployment_id) + f"{challenger_id}/"
        challenger = cls.from_location(path)
        challenger.deployment_id = deployment_id
        return challenger

    @classmethod
    def list(cls, deployment_id: str) -> List[Challenger]:
        """List all challengers for a deployment

        Parameters
        ----------
        deployment_id : str
            The ID of the deployment

        Returns
        -------
        challengers: list
            A list of challenger objects

        Examples
        --------
        .. code-block:: python

            from datarobot import Challenger
            challengers = Challenger.list(deployment_id="5c939e08962d741e34f609f0")

            challengers[0].id
            >>>'5c939e08962d741e34f609f0'
            challengers[0].model_package['name']
            >>> 'Elastic-Net Classifier'
        """
        path = cls._path.format(deployment_id)
        get_response = cls._client.get(path).json()
        challengers = [cls.from_server_data(d) for d in get_response["data"]]
        for challenger in challengers:
            challenger.deployment_id = deployment_id
        return challengers

    def delete(self) -> None:
        """Delete a challenger for a deployment"""
        path = self._path.format(self.deployment_id) + f"{self.id}/"
        self._client.delete(path)

    def update(
        self,
        name: Optional[str] = None,
        prediction_environment_id: Optional[str] = None,
    ) -> None:
        """Update name and prediction environment of a challenger

        Parameters
        ----------
        name: str, optional
            The name of the challenger model
        prediction_environment_id: str, optional
            The prediction environment id of the challenger model
        """
        payload = {}
        if name:
            payload["name"] = name
        if prediction_environment_id:
            payload["predictionEnvironmentId"] = prediction_environment_id
        if not payload:
            raise ValueError(
                "No update parameters (name and/or prediction_environment_id) provided"
            )
        path = self._path.format(self.deployment_id) + f"{self.id}/"
        response = self._client.patch(path, data=payload)
        updated_challenger = self.from_server_data(response.json())
        if name:
            self.name = updated_challenger.name
        if prediction_environment_id:
            self.prediction_environment = updated_challenger.prediction_environment



#
# Copyright 2021-2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict

import trafaret as t

from datarobot._compat import String
from datarobot.models.api_object import APIObject


class DeploymentSharedRole(APIObject):
    """
    Parameters
    ----------
    share_recipient_type: enum('user', 'group', 'organization')
        Describes the recipient type, either user, group, or organization.
    role: str, one of enum('CONSUMER', 'USER', 'OWNER')
        The role of the org/group/user on this deployment.
    id: str
        The ID of the recipient organization, group or user.
    name: string
        The name of the recipient organization, group or user.
    """

    _converter = t.Dict(
        {
            t.Key("share_recipient_type"): t.Enum("user", "group", "organization"),
            t.Key("role"): t.Enum("CONSUMER", "USER", "OWNER"),
            t.Key("id"): String(allow_blank=False, min_length=24, max_length=24),
            t.Key("name"): String(allow_blank=False),
        }
    )

    def __init__(
        self, id: str, name: str, role: str, share_recipient_type: str, **kwargs: Any
    ) -> None:
        self.share_recipient_type = share_recipient_type
        self.role = role
        self.id = id
        self.name = name

    def to_dict(self) -> Dict[str, str]:
        return {
            "role": self.role,
            "id": self.id,
            "share_recipient_type": self.share_recipient_type,
            "name": self.name,
        }


class DeploymentGrantSharedRoleWithId:
    """

    Parameters
    ----------
    share_recipient_type: enum('user', 'group', 'organization')
        Describes the recipient type, either user, group, or organization.
    role: enum('OWNER', 'USER', 'OBSERVER', 'NO_ROLE')
        The role of the recipient on this entity. One of OWNER, USER, OBSERVER, NO_ROLE.
        If NO_ROLE is specified, any existing role for the recipient will be removed.
    id: str
        The ID of the recipient.
    """

    def __init__(
        self,
        id: str,
        role: str,
        share_recipient_type: str = "user",
        **kwargs: Any,
    ) -> None:
        self.share_recipient_type = share_recipient_type
        self.role = role
        self.id = id

    def to_dict(self) -> Dict[str, str]:
        return {
            "role": self.role,
            "id": self.id,
            "share_recipient_type": self.share_recipient_type,
        }


class DeploymentGrantSharedRoleWithUsername:
    """

    Parameters
    ----------
    role: string
        The role of the recipient on this entity. One of OWNER, USER, CONSUMER, NO_ROLE.
        If NO_ROLE is specified, any existing role for the user will be removed.
    username: string
        Username of the user to update the access role for.
    """

    def __init__(
        self,
        role: str,
        username: str,
        **kwargs: Any,
    ) -> None:
        self.share_recipient_type = "user"
        self.role = role
        self.username = username

    def to_dict(self) -> Dict[str, str]:
        return {
            "role": self.role,
            "username": self.username,
            "share_recipient_type": self.share_recipient_type,
        }



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import List, Optional

import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.models.model_registry.common import UserMetadata
from datarobot.models.model_registry.registered_model_version import (
    BiasAndFairness,
    Dataset,
    ImportMeta,
    MlpkgFileContents,
    ModelDescription,
    ModelKind,
    SourceMeta,
    TagWithId,
    Target,
    Timeseries,
)


class ChampionModelPackage(APIObject):
    """
    Represents a champion model package.

    Parameters
    ----------
    id : str
        The ID of the registered model version.
    registered_model_id : str
        The ID of the parent registered model.
    registered_model_version : int
        The version of the registered model.
    name : str
        The name of the registered model version.
    model_id : str
        The ID of the model.
    model_execution_type : str
        The type of model package (version). `dedicated` (native DataRobot models) and
        custom_inference_model` (user added inference models) both execute on DataRobot
        prediction servers, while `external` does not.
    is_archived : bool
        Whether the model package (version) is permanently archived (cannot be used in deployment or
            replacement).
    import_meta : ImportMeta
        Information from when this model package (version) was first saved.
    source_meta : SourceMeta
        Meta information from where the model was generated.
    model_kind : ModelKind
        Model attribute information.
    target : Target
        Target information for the registered model version.
    model_description : ModelDescription
        Model description information.
    datasets : Dataset
        Dataset information for the registered model version.
    timeseries : Timeseries
        Time series information for the registered model version.
    bias_and_fairness : BiasAndFairness
        Bias and fairness information for the registered model version.
    is_deprecated : bool
        Whether the model package (version) is deprecated (cannot be used in deployment or
            replacement).
    build_status : str or None
        Model package (version) build status. One of `complete`, `inProgress`, `failed`.
    user_provided_id : str or None
        User provided ID for the registered model version.
    updated_at : str or None
        The time the registered model version was last updated.
    updated_by : UserMetadata or None
        The user who last updated the registered model version.
    tags : List[TagWithId] or None
        The tags associated with the registered model version.
    mlpkg_file_contents : str or None
        The contents of the model package file.
    """

    _converter = t.Dict(
        {
            t.Key("id"): t.String,
            t.Key("registered_model_id"): t.String,
            t.Key("registered_model_version"): t.Int,
            t.Key("name"): t.String,
            t.Key("model_id"): t.String,
            t.Key("model_execution_type"): t.String,
            t.Key("is_archived"): t.Bool,
            t.Key("import_meta"): t.Dict().allow_extra("*"),
            t.Key("source_meta"): t.Dict().allow_extra("*"),
            t.Key("model_kind"): t.Dict().allow_extra("*"),
            t.Key("target"): t.Dict().allow_extra("*"),
            t.Key("model_description"): t.Dict().allow_extra("*"),
            t.Key("datasets"): t.Dict().allow_extra("*"),
            t.Key("timeseries"): t.Dict().allow_extra("*"),
            t.Key("is_deprecated"): t.Bool,
            t.Key("bias_and_fairness", optional=True): t.Or(t.Dict().allow_extra("*"), t.Null),
            t.Key("build_status", optional=True): t.Or(t.String, t.Null),
            t.Key("user_provided_id", optional=True): t.Or(t.String, t.Null),
            t.Key("updated_at", optional=True): t.Or(t.String, t.Null),
            t.Key("updated_by", optional=True): t.Or(t.Dict().allow_extra("*"), t.Null),
            t.Key("tags", optional=True): t.Or(t.List(t.Dict().allow_extra("*")), t.Null),
            t.Key("mlpkg_file_contents", optional=True): t.Or(t.Dict().allow_extra("*"), t.Null),
        }
    ).allow_extra("*")

    def __init__(
        self,
        id: str,
        registered_model_id: str,
        registered_model_version: int,
        name: str,
        model_id: str,
        model_execution_type: str,
        is_archived: bool,
        import_meta: ImportMeta,
        source_meta: SourceMeta,
        model_kind: ModelKind,
        target: Target,
        model_description: ModelDescription,
        datasets: Dataset,
        timeseries: Timeseries,
        is_deprecated: bool,
        bias_and_fairness: Optional[BiasAndFairness] = None,
        build_status: Optional[str] = None,
        user_provided_id: Optional[str] = None,
        updated_at: Optional[str] = None,
        updated_by: Optional[UserMetadata] = None,
        tags: Optional[List[TagWithId]] = None,
        mlpkg_file_contents: Optional[MlpkgFileContents] = None,
    ):
        self.id = id
        self.registered_model_id = registered_model_id
        self.registered_model_version = registered_model_version
        self.name = name
        self.model_id = model_id
        self.model_execution_type = model_execution_type
        self.is_archived = is_archived
        self.import_meta = import_meta
        self.source_meta = source_meta
        self.model_kind = model_kind
        self.target = target
        self.model_description = model_description
        self.datasets = datasets
        self.timeseries = timeseries
        self.bias_and_fairness = bias_and_fairness
        self.build_status = build_status
        self.user_provided_id = user_provided_id
        self.updated_at = updated_at
        self.updated_by = updated_by
        self.tags = tags
        self.mlpkg_file_contents = mlpkg_file_contents
        self.is_deprecated = is_deprecated
